import os
import sys
import time
from collections import namedtuple
from math import ceil

import requests
from prompt_toolkit import print_formatted_text
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.shortcuts import ProgressBar

from swiggy_analytics.cli import get_input_value, print_bars, quit_prompt
from swiggy_analytics.constants import (PROGRESS_BAR_FORMATTER,
                                        PROGRESS_BAR_STYLE,
                                        SWIGGY_API_CALL_INTERVAL,
                                        SWIGGY_LOGIN_URL, SWIGGY_ORDER_URL,
                                        SWIGGY_URL)
from swiggy_analytics.db import SwiggyDB
from swiggy_analytics.exceptions import (SwiggyAPIError, SwiggyCliAuthError,
                                         SwiggyCliConfigError,
                                         SwiggyCliQuitError, SwiggyDBError)
from swiggy_analytics.queries import (get_items_name_count_query,
                                      get_monthly_spend_count,
                                      get_order_count_day_of_week,
                                      get_top_20_restaurants_query,
                                      get_total_amount_query,
                                      get_total_orders_query)
from swiggy_analytics.utils import (format_amount, get_config, get_month,
                                    get_scores,
                                    get_weekday_name, save_config)

session = requests.Session()


def fetch_orders_info(orders):
    """
    Parses the list of orders and extracts only the fields
    required for analytics. Prepares the result as
    input for the db
    """
    order_details = []
    order_items = []
    OrderDetails = namedtuple(
        'OrderDetails', ['order_id', 'order_total', 'restaurant_name', 'order_time', 'rain_mode', 'on_time'])
    OrderItems = namedtuple(
        'OrderItems', ['order_id', 'name', 'is_veg'])

    # filter orders which are delivered
    delivered_orders = list(filter(lambda i: i.get(
        'order_status', '') == 'Delivered', orders))
    for order in delivered_orders:
        order_id = order.get('order_id')
        order_total = order.get('order_total')
        restaurant_name = order.get('restaurant_name')
        order_time = order.get('order_time')
        rain_mode = order.get('rain_mode', False)
        on_time = order.get('on_time', True)

        order_details.append(OrderDetails(order_id=order_id,
                                          order_total=order_total,
                                          restaurant_name=restaurant_name,
                                          order_time=order_time,
                                          rain_mode=rain_mode,
                                          on_time=on_time))
        if order.get('order_items'):
            for item in order.get('order_items'):
                is_veg = item.get('is_veg')
                name = item.get('name')
                order_items.append(OrderItems(order_id=order_id,
                                              name=name,
                                              is_veg=is_veg
                                              ))

    return {'order_details': order_details, 'order_items': order_items}


def fetch_orders(offset_id):
    """
    Fetches a set of orders (limited to 10) using the history API
    """
    try:
        response = session.get(
            SWIGGY_ORDER_URL + '?order_id=' + str(offset_id))
    except requests.exceptions.ConnectionError:
        fetch_orders(offset_id)
    except Exception as e:
        raise SwiggyAPIError("Error while fetching orders: {}".format(e))
    return response.json().get('data').get('orders', [])


def initial_setup_prompt():
    """
    Prompt shown for the first time setup
    or when configure flag is passed.
    Fetch the keys from user and store in config file
    """
    try:
        swiggy_username = get_input_value(title='First time setup',
                                          text='Please enter your swiggy username. You can use your mobile number')
    except SwiggyCliQuitError:
        sys.exit("Bye")
    try:
        swiggy_password = get_input_value(
            title='First time setup',
            text='Please enter your swiggy password',
            password=True)
    except SwiggyCliQuitError:
        sys.exit("Bye")

    save_config(username=swiggy_username, password=swiggy_password)
    return None


def perform_login():
    """
    Attemps to make a GET request to Swiggy and on success,
    uses the CSRF token to make a POST request to login and
    maintain the cookies in the same session object, which is used
    for further calls.
    """
    establish_connection = session.get(SWIGGY_URL)
    # This is the most ugliest parsing I have ever written. Don't @ me
    csrf_token = establish_connection.text.split("csrfToken")[1].split("=")[
        1].split(";")[0][2:-1]

    if not csrf_token:
        raise SwiggyCliAuthError("Unable to parse CSRF Token. Login failed")

    # fetch username, password from config
    try:
        username, password = get_config()
    except SwiggyCliConfigError as e:
        raise e
    login_response = session.post(SWIGGY_LOGIN_URL, headers={'content-type': 'application/json',
                                                             'User-Agent': 'Mozilla/Gecko/Firefox/65.0'},
                                  json={"mobile": username, "password": password, '_csrf': csrf_token})

    if login_response.text == "Invalid Request":
        perform_login()

    if login_response.status_code != 200:
        raise SwiggyCliAuthError(
            "Login response non success {}".format(login_response.status_code))


def insert_orders_data(db, orders):
    # extract only the fields required for analytics
    try:
        orders_info = fetch_orders_info(orders)
    except SwiggyAPIError as e:
        raise SwiggyAPIError(e)

    # store the order data in db
    try:
        db.insert_orders_details(orders_info.get('order_details'))
    except SwiggyDBError as e:
        print(e)
    try:
        db.insert_order_items(orders_info.get('order_items'))
    except SwiggyDBError as e:
        print(e)


def fetch_and_store_orders(db):
    """
    Fetches all the historical orders for the user and saves them in db
    """
    response = session.get(SWIGGY_ORDER_URL)
    if not response.json().get('data', None):
        raise SwiggyAPIError("Unable to fetch orders")

    # get the last order_id to use as offset param for next order fetch call
    orders = response.json().get('data').get('orders', None)
    if not orders:
        raise SwiggyAPIError("Unable to fetch orders")

    # extract order meta data and insert in db
    insert_orders_data(db, orders)

    offset_id = orders[-1]['order_id']
    count = response.json().get('data')['total_orders']
    pages = ceil(count/10)
    label = "Fetching {} orders".format(count)

    # Updates the progress bar on every orders fetch call (i.e. after 10 unique orders)
    with ProgressBar(style=PROGRESS_BAR_STYLE, formatters=PROGRESS_BAR_FORMATTER) as pb:
        for i in pb(range(pages), label=label):
            try:
                orders = fetch_orders(offset_id)
            except SwiggyAPIError as e:
                raise SwiggyAPIError(e)
            if len(orders) == 0:
                break

            # extract order meta data and insert in db
            insert_orders_data(db, orders)
            # Responsible Scraping. Code word for "dont wanna overload their servers :P" :)

            time.sleep(SWIGGY_API_CALL_INTERVAL)
            # SAD PANDA FACE BEGIN
            # The way it works is that, the first API call returns a paginated set of 10 orders and to fetch the next result, you need
            # to send the last order_id from this result set as an offset parameter. Because the way this offset/cursor
            # is designed it makes it impossible to use any kind of async/await magic.
            # SAD PANDA FACE OVER
            offset_id = orders[-1]['order_id']


def display_stats(db):
    """
    Queries the DB to get basic stats and display
    using `print_bars`
    """
    print_formatted_text(
        HTML("Some basic stats based on your order history:"))
    # orders summary sec
    try:
        orders_count = db.fetch_result(query=get_total_orders_query)[0][0]
    except SwiggyDBError as e:
        raise("Error while fetching total orders count {}".format(e))
    print_formatted_text(HTML(
        '\nYour total <b>delivered</b> orders are: <skyblue>{}</skyblue>\n'.format(orders_count)))

    try:
        total_amount = db.fetch_result(query=get_total_amount_query)[0][0]
    except SwiggyDBError as e:
        raise("Error while fetching total amount {}".format(e))
    print_formatted_text(HTML(
        'You have spent a total sum of <skyblue>{}</skyblue>'.format(format_amount(total_amount))))

    # spend pattern
    print_formatted_text(
        HTML('\nYour spend and orders count distributed monthly:'))
    try:
        items_count_bar_graph = db.fetch_result(
            query=get_monthly_spend_count)
    except SwiggyDBError as e:
        raise("Error while fetching items v/s count {}".format(e))
    print_bars(get_scores([{"name": get_month(i[0]), "count":i[2], "extra":format_amount(i[1])}
                           for i in items_count_bar_graph]))

    # topK weekdays
    print_formatted_text(HTML(
        '\nWeekday wise distribution of your orders:'))
    try:
        weekday_count_bar_graph = db.fetch_result(
            query=get_order_count_day_of_week)
    except SwiggyDBError as e:
        raise("Error while fetching total orders v/s weekday data {}".format(e))

    print_bars(get_scores([{"name": get_weekday_name(i[0]), "count":i[1]}
                           for i in weekday_count_bar_graph]))

    # topK restaraunts
    print_formatted_text(
        HTML('\nTop <b>10</b> restaraunts from where you have ordered:'))
    try:
        items_count_bar_graph = db.fetch_result(
            query=get_top_20_restaurants_query)
    except SwiggyDBError as e:
        raise("Error while fetching items v/s count {}".format(e))
    print_bars(get_scores([{"name": i[0], "count":i[1]}
                           for i in items_count_bar_graph]))
    # topK items
    print_formatted_text(
        HTML('\nTop <b>10</b> items that you have ordered:'))
    try:
        items_count_bar_graph = db.fetch_result(
            query=get_items_name_count_query)
    except SwiggyDBError as e:
        raise("Error while fetching items v/s count {}".format(e))
    print_bars(get_scores([{"name": i[0], "count":i[1]}
                           for i in items_count_bar_graph]))



import calendar
import configparser
import os
from datetime import datetime

from swiggy_analytics.constants import CONFIG_FILEPATH


def save_config(username, password):
    """
    Saves a config file to the user's specified location.
    """
    Config = configparser.RawConfigParser()
    with open(CONFIG_FILEPATH, 'w') as config_file:
        # add the settings to the structure of the file, and lets write it out...
        Config.add_section('Auth')
        Config.set('Auth', 'Username', username)
        Config.set('Auth', 'Password', password)
        Config.write(config_file)


def get_config():
    """
    Returns the user's credentials from the config if it exists.
    """
    if not config_file_present():
        raise SwiggyCliConfigError("No config file present")

    Config = configparser.RawConfigParser()
    Config.read(CONFIG_FILEPATH)
    return Config.get('Auth', 'Username'), Config.get('Auth', 'Password')


def config_file_present():
    """
    Checks whether the config file exists or not.
    """
    return os.path.exists(os.path.join(CONFIG_FILEPATH))


def normalize(x, xmin, xmax):
    """Normalize a number to a 0-1 range given a min and max of its set."""
    return float(x - xmin) / float(xmax - xmin)


def get_scores(items):
    """Compute normalized scores (0-1) for order count."""
    vals = [i["count"] for i in items]
    vals.append(0)

    xmin = min(vals)
    xmax = max(vals)

    # Normalize.
    for i in items:
        i["score"] = normalize(i["count"], xmin, xmax)

    return items


def get_weekday_name(int_day):
    """
    Returns name of the day based on it's int representation (eg Sunday=0, Monday=1).
    """
    return calendar.day_name[int_day-1]


def get_month(date):
    """
    Converts string representation of date in form of 2018-11-10 as Nov-18
    """
    return datetime.strptime(date, '%Y-%m-%d').strftime('%b-%y')


def format_amount(data):
    """
    Prefixes the amount with the rupee symbol.
    """
    return "₹"+str(data)



# create table queries
create_orders_table_query = """
CREATE TABLE IF NOT EXISTS orders(id INTEGER PRIMARY KEY AUTOINCREMENT,
order_id INT UNIQUE, order_total FLOAT, restaurant_name VARCHAR,
order_time DATETIME, rain_mode BOOLEAN, on_time BOOLEAN)
"""

create_items_table_query = """
CREATE TABLE IF NOT EXISTS items(id INTEGER PRIMARY KEY AUTOINCREMENT, order_id INT,
name VARCHAR, is_veg BOOLEAN,
FOREIGN KEY(order_id) REFERENCES orders(order_id))
"""

# insert records queries
insert_orders_query = """
INSERT INTO orders(order_id, order_total, restaurant_name, order_time, rain_mode, on_time) VALUES (?,?,?,?,?,?)
"""
insert_items_query = """
INSERT INTO items(order_id, name, is_veg) VALUES (?,?,?)
"""

# analytics queries
get_total_orders_query = """
SELECT count(order_id) from orders
"""
get_total_amount_query = """
SELECT sum(order_total) from orders
"""
get_items_name_count_query = """
SELECT name, count(id) as ctr from items group by name order by ctr desc limit 10
"""
get_top_20_restaurants_query = """
SELECT "restaurant_name" AS "restaurant_name", count(*) AS "count"
FROM "orders"
GROUP BY "restaurant_name"
ORDER BY "count" DESC LIMIT 10
"""
get_order_count_day_of_week = """
SELECT CAST((strftime('%w', "order_time")) AS integer) AS "order_time", count(*) AS "count"
FROM "orders"
GROUP BY CAST((strftime('%w', "order_time")) AS integer)
ORDER BY CAST((strftime('%w', "order_time")) AS integer) ASC
"""
get_monthly_spend_count = """
SELECT date("order_time", 'start of month') AS "order_time", sum("order_total") AS "sum", count(*) AS "count"
FROM "orders"
GROUP BY date("order_time", 'start of month')
ORDER BY date("order_time", 'start of month') ASC
"""



import sys

from prompt_toolkit import prompt
from prompt_toolkit.completion import WordCompleter
from prompt_toolkit.shortcuts import button_dialog, input_dialog
from prompt_toolkit.validation import ValidationError, Validator

from swiggy_analytics.constants import NO_ANSWER_CHOICES, YES_ANSWER_CHOICES
from swiggy_analytics.exceptions import SwiggyCliQuitError


class YesNoValidator(Validator):

    def validate(self, document):
        """
        Display error message if text not valid.

        return:
            A status bar with validation error message
        """
        text = document.text.lower()
        if text not in YES_ANSWER_CHOICES+NO_ANSWER_CHOICES:
            raise ValidationError(message='Enter "yes" if you want to continue '
                                          '"no" if you want to refetch the orders.')


def quit_prompt():
    """
    Generic function to display a confirmation screen shown if
    user decides to cancel any operation.
    """
    result = button_dialog(
        title='Are you sure you want to exit?',
        text='Do you want to confirm?',
        buttons=[
            ('Yes', True),
            ('No', False),
        ],
    )
    return result


def get_input_value(title, text, password=False):
    """
    Handle input by user
    """
    value = input_dialog(title, text, password=password)
    if not value:
        # if user decides to cancel
        if not quit_prompt():
            # show the input dialogue again if user doesn't want to quit
            return input_dialog(title, text, password=password)
        # in case user wants to cancel
        raise SwiggyCliQuitError("No input provided by user")
    # raw value entered by user
    return value


# SOURCE: Modified version of https://github.com/knadh/git-bars/blob/master/gitbars/gitbars.py#L24
def print_bars(items=None, block=u"\u2580", width=50):
    """
    Print unicode bar representations of dates and scores.
    """
    for i in range(len(items)):
        num = str(items[i]["count"])

        sys.stdout.write(" ")
        sys.stdout.write(num)
        sys.stdout.write((5 - len(num)) * " ")

        # Colour the weekend bars.
        # if items[i]["weekend"]:
        #     sys.stdout.write("\033[93m")

        sys.stdout.write(block * int(items[i]["score"] * width))

        sys.stdout.write("\033[93m")
        sys.stdout.write(" {} ".format(items[i]["name"]))
        sys.stdout.write("\x1b[0m")

        if items[i].get("extra"):
            sys.stdout.write("\033[91m")
            sys.stdout.write((items[i].get("extra")))
            sys.stdout.write("\x1b[0m")

        sys.stdout.write("\n")


def user_continue():
    """
    Validates the user's answer as one of Yes/No
    """
    html_completer = WordCompleter(YES_ANSWER_CHOICES+NO_ANSWER_CHOICES)
    answer = prompt('You already have a swiggy.db file, do you want to use the same? (Yes/no) ', completer=html_completer,
                    validator=YesNoValidator(), default="yes")

    if answer.lower() in YES_ANSWER_CHOICES:
        return True
    return False



class SwiggyCliQuitError(Exception):
    pass


class SwiggyCliAuthError(Exception):
    pass


class SwiggyCliConfigError(Exception):
    pass


class SwiggyAPIError(Exception):
    pass


class SwiggyDBError(Exception):
    pass



import os
from pathlib import Path

from prompt_toolkit.shortcuts.progress_bar import formatters
from prompt_toolkit.styles import Style

SWIGGY_URL = 'https://www.swiggy.com'
SWIGGY_LOGIN_URL = SWIGGY_URL + '/dapi/auth/signin-with-check'
SWIGGY_ORDER_URL = SWIGGY_URL + '/dapi/order/all'
SWIGGY_API_CALL_INTERVAL = 1.5  # interval between API calls. (in seconds)

CONFIG_FILEPATH = os.path.join(
    str(Path.home()), '.swiggy-analytics-config.ini')
DB_FILEPATH = os.path.join(str(os.getcwd()), 'swiggy.db')

PROGRESS_BAR_STYLE = Style.from_dict({
    'label': 'bg:#FFA500 #000000',
    'percentage': 'bg:#FFA500 #000000',
    'current': '#448844',
    'bar': '',
})

PROGRESS_BAR_FORMATTER = [
    formatters.Label(),
    formatters.Text(': [', style='class:percentage'),
    formatters.Percentage(),
    formatters.Text(']', style='class:percentage'),
    formatters.Text(' '),
    formatters.Bar(sym_a='#', sym_b='#', sym_c='.'),
    formatters.Text('  '),
]

YES_ANSWER_CHOICES = ['y', 'yes', 'yeah', 'yup']
NO_ANSWER_CHOICES = ['n', 'no', 'nope']



import sqlite3

from swiggy_analytics.constants import DB_FILEPATH
from swiggy_analytics.exceptions import SwiggyDBError
from swiggy_analytics.queries import (create_items_table_query,
                                      create_orders_table_query,
                                      insert_items_query, insert_orders_query)


class SwiggyDB(object):
    def init_db(self, persist=False):
        try:
            if persist:
                print("Connecting to {}".format(DB_FILEPATH))
                self.conn = sqlite3.connect(DB_FILEPATH)
            else:
                print("Connecting to in memory db")
                self.conn = sqlite3.connect(':memory:')
        except Exception:
            raise SwiggyCliDBException("Unable to connect to DB")

    def create_db(self):
        cur = self.conn.cursor()
        cur.execute(create_orders_table_query)
        cur.execute(create_items_table_query)
        self.conn.commit()

    def insert_orders_details(self, orders):
        cur = self.conn.cursor()
        # CAVEAT: Since this is a batch insert, even if one of the order already exists, the whole
        # transaction will fail and result in not adding the other unique orders
        # This is by design. Will think of a better way to handle later, it is not a huge issue now.
        try:
            cur.executemany(insert_orders_query, orders)
            self.conn.commit()
        except sqlite3.Error as e:
            if "UNIQUE" in "{}".format(e):
                pass
            else:
                raise SwiggyDBError(
                    "Error while inserting orders: {}".format(e))
        except Exception as e:
            raise SwiggyDBError("Error while executing query: {}".format(e))

    def insert_order_items(self, items):
        cur = self.conn.cursor()
        try:
            cur.executemany(insert_items_query, items)
            self.conn.commit()
        except sqlite3.Error as e:
            if "UNIQUE" in "{}".format(e):
                pass
            else:
                raise SwiggyDBError("Error while inserting items {}".format(e))
        except Exception as e:
            raise SwiggyDBError("Error while executing query: {}".format(e))

    def fetch_result(self, query):
        cur = self.conn.cursor()
        try:
            cur.execute(query)
        except sqlite3.Error as e:
            raise SwiggyDBError("Error while fetching items: {}".format(e))
        except Exception as e:
            raise SwiggyDBError("Error while executing query: {}".format(e))
        return cur.fetchall()




#!/usr/bin/python

"""
swiggy-cli-expense is a utility to fetch your orders
history from Swiggy and provide basic stats on them.

Author: Karan Sharma, http://mrkaran.com

Licensed under the MIT License.
"""

import argparse
import os
import sys

from swiggy_analytics.cli import user_continue
from swiggy_analytics.constants import DB_FILEPATH
from swiggy_analytics.db import SwiggyDB
from swiggy_analytics.exceptions import (SwiggyAPIError, SwiggyCliAuthError,
                                         SwiggyCliConfigError)
from swiggy_analytics.helper import (display_stats, fetch_and_store_orders,
                                     initial_setup_prompt, perform_login)
from swiggy_analytics.utils import config_file_present


def main():
    parser = argparse.ArgumentParser(
        description="Fetch your past swiggy orders " +
                    "using the command line",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument('--configure', action='store_true',
                        help="Configure  swiggy-analytics  CLI  options. On running this command, " +
                        "you will be prompted for configuration values such as your Swiggy Username " +
                        "and your Swiggy Password. " +
                        "(the default location is ~/.aws/config). The AWS CLI will create/overwrite " +
                        "a configuration file for you, at ~/.swiggy-exepnse-config. \n" +

                        "Note: If you are not comfortable sharing your username and password with the utility, " +
                        "you can audit the code yourself. This file will only live in your filesystem and " +
                        "accessible by your username only. "
                        )
    parser.add_argument('--save', action='store_true',
                        help="Use this flag if you want to store your orders " +
                        "in a sqllite db file. After the program has completed the " +
                        "fetching orders, `swiggy.db` file wile be created in you current working directory. ")

    args = parser.parse_args()

    print('''Welcome to swiggy-analytics.

This command line tool will help you fetch your order history from https://swiggy.com.
You can choose to persist the detailed order information in a SQLite database or
perform lightweight stats operations using in-memory calculations.
''')
    if not config_file_present() or args.configure:
        initial_setup_prompt()

    if os.path.exists(DB_FILEPATH) and user_continue():
        db = SwiggyDB()
        # connect to the existing db
        db.init_db(persist=True)
        # show basic stats by fetching the results from already existing db
        display_stats(db)
        return None

    db = SwiggyDB()
    db.init_db(persist=args.save)
    db.create_db()

    try:
        perform_login()
    except SwiggyCliConfigError:
        sys.exit(
            "Error reading config file. Please generate a config file using --configure flag.")
    except SwiggyCliAuthError:
        sys.exit("Login to swiggy failed.")

    try:
        orders = fetch_and_store_orders(db)
    except SwiggyAPIError:
        sys.exit(
            "Error fetching orders from Swiggy. " +
            "Please check your credentials. " +
            "You can use swiggy-analytics --configure to regenerate")

    display_stats(db)

    return None


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python

from os import path

from setuptools import find_packages, setup


# read the contents of your README file
def read_readme():
    this_directory = path.abspath(path.dirname(__file__))
    with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
        return f.read()


def requirements():
    with open('requirements.txt') as f:
        return f.read().splitlines()


setup(
    name="swiggy_analytics",
    version="1.5",
    description="A CLI for fetching past orders from Swiggy and performing basic stats in the terminal",
    long_description=read_readme(),
    # long_description_content_type='text/markdown',
    author="Karan Sharma",
    author_email="karansharma1295@gmail.com",
    url="https://github.com/mr-karan/swiggy-analytics",
    install_requires=requirements(),
    include_package_data=True,
    packages=find_packages(),
    download_url="https://github.com/mr-karan/swiggy-analytics",
    license="MIT License",
    entry_points={
        'console_scripts': [
            'swiggy-analytics = swiggy_analytics.swiggy_analytics:main',
        ],
    },
    classifiers=[
        "Topic :: Utilities",
        "Intended Audience :: End Users/Desktop",
        "Programming Language :: Python",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        'Topic :: Software Development :: Libraries',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3 :: Only',
    ]
)
